#
# Sensus Consulting Ltd (C) 1997
# Matt Newman <matt@sensus.org>
#
# TODO:
#	1.	on unexpand, if selection is below, move up tree.
#	2.	context requests
#	3.	drag&drop
#
#
#INDEX\
proc tk++::TreeView {} {}

option add *TreeView.padX 18 widgetDefault
option add *TreeView.padY 16 widgetDefault
option add *TreeView.background white widgetDefault
option add *TreeView.highlightThickness 0 widgetDefault
option add *TreeView.width 1 widgetDefault
option add *TreeView.height 1 widgetDefault

package require tk++

proc treeview {w args} {
	return [uplevel 1 tk++::TreeView $w $args]
}

class tk++::TreeView {
	inherit itk::Widget

	constructor {args} {}

	# Tree methods
	public {
		method add {o args}
		method children {o args}
		method delete {o}
		method export {{o 0}}
		method find {o fld val {defval ""}}
		method get {o ?option? ?default?}
		method import {o data}
		method insert {o idx args}
		method isdesendant {o p}
		method parent {o}
		method path {o {sep /}}
		method put {o args}

		method sort {list args} {}

		# TreeView methods
		method activate {o}
		method closest {x y}
		method context {x y}
		method curselection {}
		method expand {o}
		method isvisable {o}
		method isexpanded {o}
		method see {o}
		method select {x y}
		method toggle {x y}
		method unexpand {o}
	}

	itk_option define -background background Background white
	itk_option define -highlightthickness highlightThickness HighlightThickness 0 
	itk_option define -height height Height 1
	itk_option define -width width Width 1

	itk_option define -command command Command ""
	itk_option define -controls controls Controls [list [img plus] [img minus]]
	# Hide the root plus/minus control.
	itk_option define -hide hide Hide 1
	itk_option define -padx padX PadX 18
	itk_option define -pady padY PadY 16

	protected {
		method sortPair {flds id1 id2} {}
		method callback {o args}
		method _see {view size t0 t1 c0 c1}
		method drawNode {o {xc -1000} {yc -1000}}
		method newNode {po o data}
		method refresh {o}
		method refreshNode {o}

		variable uid 0
		variable selection {}
		variable store
	}
}

body tk++::TreeView::constructor {args} {

	itk_component add scroll {
		scroll $itk_interior.c canvas -relief flat -borderwidth 0
	} {
		keep -relief -borderwidth
	}
	itk_component add canvas {
		$itk_component(scroll) child
	} {
		keep -background -highlightthickness
		keep -width -height -xscrollincrement -yscrollincrement
		rename -xscrollincrement -padx padX PadX
		rename -yscrollincrement -pady padY PadY
	}
	pack $itk_component(scroll) -anchor nw -side left -fill both -expand yes

	# Bindings
	$itk_component(canvas) bind ctrl	<Button-1>		\
										[code $this toggle %x %y]
	$itk_component(canvas) bind folder	<Button-1>		\
										[code $this select %x %y]
	$itk_component(canvas) bind folder	<Double-Button-1>	\
										[code $this toggle %x %y]
	$itk_component(canvas) bind folder	<Button-3>		\
										[code $this context %x %y]

	$itk_component(canvas) conf -scrollregion {0 0 1 1}
	$itk_component(canvas) xview moveto 0.0
	$itk_component(canvas) yview moveto 0.0

	#
	# Setup 
	#
	eval itk_initialize $args

	newNode 0 0 [list image folder]
	expand 0

	$itk_component(canvas) xview moveto 0.0
	$itk_component(canvas) yview moveto 0.0
}

body tk++::TreeView::closest {x y} {
	set tag [$itk_component(canvas) find closest \
						[$itk_component(canvas) canvasx $x] \
						[$itk_component(canvas) canvasy $y] ]
	if {$tag == ""} return

	# 1st tag is always object
	set o [lindex [$itk_component(canvas) itemcget $tag -tags] 0]
	if {![string match *.all $o]} {
		error "unexpected tag: \"$o\""
	}
	regsub {\.all$} $o {} o
	return $o
}
#
# Conext Request: typically client of this widget will post
#		a "context" menu of operations that may be
#		performed on this node.
#
body tk++::TreeView::context {x y} {
	set o [closest $x $y]
	if {$o == ""} return

	activate $o

	# WARNING:	re-entrant
	# This callback is expected to invoke
	# our insert/delete methods
	callback $o context $x $y
}
body tk++::TreeView::curselection {} {
	return $selection
}
#
# Selection:
#
body tk++::TreeView::select {x y} {
	set o [closest $x $y]
	if {$o == ""} return

	activate $o
}
body tk++::TreeView::toggle {x y} {
	set o [closest $x $y]
	if {$o == ""} return

	if {$store($o,expanded)} {
		unexpand $o
	} else {
		expand $o
	}
}
#
# Node specific methods
#
body tk++::TreeView::activate {o} {

	if {$o != "none" && $store($o,selected)} return

	if {$selection != ""} {
		callback $selection activate 0
		$itk_component(canvas) select clear
		set store($selection,selected)	0
		set selection {}
	}
	if {$o == "none"} return

	callback $o activate 1

	set selection	$o
	set store($o,selected)	1

	refreshNode $o

	see $o
}
body tk++::TreeView::add {o args} {
	set po $o
	set o [incr uid]

	lappend store($po,children)	$o
	newNode $po $o $args

	return $o
}
body tk++::TreeView::children {o args} {
	if {[llength $args]==0} {
		return $store($o,children)
	}
	set ret {}
	foreach cid $store($o,children) {
		set ok 1
		foreach {fld val} $args {
			set cval [pget $store($cid) $fld __NULL__]
			if {[string compare $val $cval]!=0}	{
				set ok 0
				break
			}
		}
		if {$ok} {
			lappend ret $cid
		}
	}
	return $ret
}
body tk++::TreeView::delete {o {master 1}} {

	if {$master && [isdesendant $selection $o]} {
		activate $store($o,parent)
	}

	foreach co $store($o,children) {
		delete $co 0
	}
	if {$o == 0} {
		set store(0,children) {}
		refresh 0
		return
	}
	# remove canvas objects
	$itk_component(canvas) delete $o.all

	set po $store($o,parent)
	# reclaim storage
	unset store($o,parent)
	unset store($o,children)
	unset store($o,tags)
	unset store($o,expanded)
	unset store($o,selected)
	unset store($o,h)

	if $master {
		set idx [lsearch $store($po,children) $o]
		set store($po,children)	[lreplace $store($po,children) $idx $idx]

		# Visable & expanded
		if {[isvisable $po]} {
			refresh $po
		}
	}
	return ""
}
#
# Expand nodes immediately below $o
#
body tk++::TreeView::expand {o} {
	set xy [$itk_component(canvas) coords $o.image]
	if {$xy == ""} {
		expand $store($o,parent)
	}
	if $store($o,expanded)	return

	# WARNING:	re-entrant
	# This callback is expected to invoke
	# our insert/delete methods
	callback $o expand 1

	set store($o,expanded)	1
	foreach co $store($o,children) {
		drawNode $co
	}
	refresh $o

	see $o

	return $store($o,children)
}
body tk++::TreeView::find {o fld val {defval ""}} {
	foreach co $store($o,children) {
		if {[string compare $val [pget $store($co) $fld $defval]]==0} {
			return $co
		}
	}
	return -1
}
body tk++::TreeView::get {o {option ""} {defval "!@#%"}} {
	if {$option == ""} {
		return $store($o)
	} elseif {[string compare $defval "!@#%"]==0} {
		return [pget $store($o) $option]
	} else {
		if {[catch {pget $store($o) $option} ret]} {
			set ret $defval
		}
		return $ret
	}
}
body tk++::TreeView::isdesendant {o p} {
	if {$o == ""} {
		return 0
	}
	if {[lsearch $store($o,tags) $p.all]!=-1} {
		return 1
	} else {
		return 0
	}
}
body tk++::TreeView::isvisable {o} {
	set po $store($o,parent)
	if {$o == 0 || \
		($store($po,expanded) && [$itk_component(canvas) coords $po.image] != "")} {
		return 1
	} else {
		return 0
	}
}
body tk++::TreeView::isexpanded {o} {
	return $store($o,expanded)
}
#
# Create a new node, as a child of $po
#
body tk++::TreeView::insert {o idx args} {
	set po $o
	set o [incr uid]

	set store($po,children)	[linsert $store($po,children) $idx $o]
	newNode $po $o $args

	return $o
}
body tk++::TreeView::parent {o} {
	return $store($o,parent)
}
body tk++::TreeView::path {o {sep /}} {
	set name [pget $store($o) name]
	set po $store($o,parent)

	if {$po == $o} {
		return ""
	} else {
		set path [path $po $sep]
		if {$path == ""} {
			return $name
		} else {
			return "$path$sep$name"
		}
	}
}
body tk++::TreeView::put {o args} {
	set store($o) [eval [list pset $store($o)] $args]
	set store($o) [pset $store($o) image [img [pget $store($o) image]]]
	refreshNode $o
}
body tk++::TreeView::see {o} {
	lassign [$itk_component(canvas) bbox all] x0 y0 x1 y1
	lassign [$itk_component(canvas) bbox $o.all] cx0 cy0 cx1 cy1
	set rw [winfo width $itk_component(canvas)]
	set rh [winfo height $itk_component(canvas)]

	_see yview $rh $y0 $y1 $cy0 $cy1
	_see xview $rw $x0 $x1 $cx0 $cx1
}
body tk++::TreeView::unexpand {o} {
	if !$store($o,expanded)	return

	if {[isdesendant $selection $o]} {
		activate $o
	}
	# WARNING:	re-entrant
	# This callback is expected to invoke
	# our insert/delete methods
	callback $o expand 0

	set store($o,expanded)	0
	set store($o,h)			$itk_option(-pady)
	foreach co $store($o,children) {
		$itk_component(canvas) delete $co.all
	}
	refreshNode $o

	lassign [$itk_component(canvas) coords $o.image] xc yc
	$itk_component(canvas) coords $o.vline $xc $yc $xc $yc

	refresh $store($o,parent)

	see $o
}
#
# Internal Routines
#
body tk++::TreeView::refreshNode {o} {
	set name [pget $store($o) name]
	set image [pget $store($o) image]

	$itk_component(canvas) itemconf $o.name -text $name
	$itk_component(canvas) itemconf $o.image -image $image

	if {$store($o,expanded)} {
		if {$store($o,children) != ""} {
			$itk_component(canvas) itemconf $o.ctrl -image [lindex $itk_option(-controls) 1]
		} else {
			$itk_component(canvas) itemconf $o.ctrl -image ""
		}
	} else {
		$itk_component(canvas) itemconf $o.ctrl -image [lindex $itk_option(-controls) 0]
	}

	if $store($o,selected) {
		catch {
			$itk_component(canvas) select from $o.name 0
			$itk_component(canvas) select to $o.name end
		}
	}
}
body tk++::TreeView::callback {o args} {
	if {[string compare $itk_option(-command) ""]==0} return

	$itk_component(canvas) conf -cursor watch
	if [catch [concat $itk_option(-command) $o $args] ret] {
		global errorInfo
		puts stderr "Background error: $errorInfo"
	}
	$itk_component(canvas) conf -cursor {}
}
body tk++::TreeView::_see {view size t0 t1 c0 c1} {
	lassign [$itk_component(canvas) $view] pv0 pv1
	set dt	[expr double($t1 - $t0)]
	set pc0	[expr ($c0 - $t0)/$dt]
	set pc1	[expr ($c1 - $t0)/$dt]

	if {$dt < $size} {
		# Display area is smaller than window
		$itk_component(canvas) $view moveto 0.0
	} elseif {$pc0 > $pv0 && $pc1 < $pv1} {
		# donothing - fits w/o scrolling
	} elseif {($pc1 - $pc0) > ($pv1 - $pv0)} {
		# Larger than window
		$itk_component(canvas) $view moveto $pc0
	} else {
		$itk_component(canvas) $view moveto [expr $pv0 + ($pc1 - $pv1)] 
	}
}
#
# Refresh - reposition children to "correct" x,y locations
#
# WARNING:	only call for expanded nodes.
#
body tk++::TreeView::refresh {o} {
	if {$o == 0} {
		set xc	[expr $itk_option(-padx)/2]
		set yc	[expr $itk_option(-pady)/2]

		set xy	[$itk_component(canvas) coords $o.image]
		set xc1	[lindex $xy 0]
		set yc1	[lindex $xy 1]
		if {$xc != $xc1 || $yc != $yc1} {
			$itk_component(canvas) move $o.all [expr $xc - $xc1] [expr $yc - $yc1]
		}
	} else {
		set xy	[$itk_component(canvas) coords $o.image]
		set xc	[lindex $xy 0]
		set yc	[lindex $xy 1]
	}
	set h0	$itk_option(-pady)
	set x0	[expr $xc + $itk_option(-padx)]
	set y0	$yc
	if $store($o,expanded) {
		foreach co $store($o,children) {
			set y0	[expr $y0 + $h0]
			set xy	[$itk_component(canvas) coords $co.image]
			set xc1	[lindex $xy 0]
			set yc1	[lindex $xy 1]

			# do a delta-move to correct location
			$itk_component(canvas) move $co.all [expr $x0 - $xc1] [expr $y0 - $yc1]
#tclLog "move $co -> $x0, $y0"

			set h0 $store($co,h)
		}
	}
	refreshNode $o

	$itk_component(canvas) coords $o.vline $xc $yc $xc $y0
	set store($o,h)		[expr ($y0 - $yc) + $h0]

	if {$o == 0} {
		$itk_component(canvas) conf -scrollregion [$itk_component(canvas) bbox all]
	} else {
		refresh $store($o,parent)
	}
}
#
# Design Philosophy
#
# 3 main types of change:
#	a.	visable structural change
#		i.	nodes added
#		ii.	nodes removed
#		iii.	nodes (un)expanded
#	b.	visable non-structural change
#		i.	image/name changed
#		ii.	control changed (plus/minus etc.)
#	c.	non-visable change - i.e. unexpanded nodes
#
# Responses:
#	a.	i.	defered
#		ii.	defered
#		iii.	inline
#	b.	inline (direct -> screen)
#	c.	inline
body tk++::TreeView::newNode {po o data} {
	if {[catch {pget $data name} name]} {
		set name node$o
	}
	if {[catch {pget $data image} image]} {
		set image [pget $store($po) image]
		regsub -- {-selected$} $image {} image
	}
	# Ensure image is loaded/cached
	set image [img $image]

	set store($o,parent)	$po
	set store($o,children)	{}
	if {[info exists store($po,tags)]} {
		set store($o,tags)	$store($po,tags)
		lappend store($o,tags)	$po.all
	} else {
		set store($o,tags)	{}
	}
	set store($o,expanded)	0
	set store($o,selected)	0
	set store($o,h)			-1

	set store($o)	[pset $data name $name image $image]

	# Visable & expanded
	if {[isvisable $o]} {
		# draw it "off screen"
		# the refresh to move it to correct location 		
		drawNode $o
	
		refresh $po
	}
	return $o
}
#
# drawNode - create canvas items for node,
#	- by default far off screen, as refresh will reposition
#	them correctly.
#
body tk++::TreeView::drawNode {o {xc -1000} {yc -1000}} {
	if {$o == 0 && $itk_option(-hide)} {
		set x	$xc
		set y	$yc
	} else {
		set x	[expr $xc - $itk_option(-padx)]
		set y	$yc
		$itk_component(canvas) create image $x $y \
				-anchor c \
				-tag [concat $o.all $o.ctrl ctrl $store($o,tags)]
		$itk_component(canvas) create line $x $y $xc $yc \
				-fill grey \
				-tag [concat $o.all $o.hline $store($o,tags)]
		$itk_component(canvas) lower $o.hline
	}
	$itk_component(canvas) create image $xc $yc \
			-anchor c -image [pget $store($o) image] \
			-tag [concat $o.all $o.image folder $store($o,tags)]
	$itk_component(canvas) create text [expr $xc + (0.75 * $itk_option(-padx))] $yc \
			-anchor w \
			-font {{MS Sans Serif} 8} -text [pget $store($o) name] \
			-tag [concat $o.all $o.name folder $store($o,tags)]
	$itk_component(canvas) create line $xc $yc $xc $yc \
			-fill grey \
			-tags [concat $o.all $o.vline $store($o,tags)]
	$itk_component(canvas) lower $o.vline

	# (un)map children
	set y0	$yc
	set h0	$itk_option(-pady)
	if $store($o,expanded) {
		foreach co $store($o,children) {
			set y0	[expr $y0 + $h0]
			set h0	[drawNode $co [expr $xc + $itk_option(-padx)] $y0]
		}
		$itk_component(canvas) coords $o.vline $xc $yc $xc $y0
	}
	refreshNode $o

	set h0	[expr ($y0 - $yc) + $h0]
	return [set store($o,h) $h0]
}
#
# Import and Export
#
body tk++::TreeView::export {o} {
	set data {}
	lappend data [concat [list node $store($o,parent) $o] $store($o)]
	foreach co $store($o,children) {
		lvarcat data [export $co]
	}
	return $data
}
body tk++::TreeView::import {o data} {
	set root [lindex $data 0]
	set map([lindex $root 2]) $o
	eval put $o [lrange $root 3 end]

	set n 1
	foreach node [lrange $data 1 end] {
		set po [lindex $node 1]
		set co [lindex $node 2]

		set ro [eval add $map($po) [lrange $node 3 end]]
		set map($co) $ro
		incr n
	}
	refresh $o
	return $n
}
body tk++::TreeView::sort {list args} {
	if {$list == "node"} {
		set id [lindex $args 0]
		set args [lrange $args 1 end]
		set list $store($id,children)
		set store($id,children) [lsort -command [code $this sortPair $args] $list]
		refresh $id
	} else {
		return [lsort -command [code $this sortPair $args] $list]
	}
}
body tk++::TreeView::sortPair {flds id1 id2} {
	set m(asc) 1
	set m(desc) -1
	set dir asc
	foreach fld $flds {
		if {[llength $fld]>1} {
			set dir [lindex $fld 0]
			set fld [lindex $fld 1]
		}
		set v1 [pget $store($id1) $fld ""]
		set v2 [pget $store($id2) $fld ""]
		if {$v1 < $v2} {
			return [expr {$m($dir) * -1}]
		} elseif {$v1 > $v2} {
			return [expr {$m($dir) * 1}]
		}
	}
	return 0
}
